@@ -11,7 +11,7 @@ use Scalar::Util;
 use Scope::Guard;
 use Storable;
 
-our $VERSION = '2.43';
+our $VERSION = '2.50';
 
 XSLoader::load 'autobox', $VERSION;
 
@@ -23,31 +23,38 @@ my $CLASS_CACHE    = {}; # reuse the same synthetic class if the type/superclass
 
 # create a shim class - actual methods are implemented by the classes in its @ISA
 #
-# as an optimization, or at least a courtesy, return the previously-generated class
+# as an optimization, return the previously-generated class
 # if we've seen the same (canonicalized) type/isa combination before
 
-sub _generate_class($@) {
+sub _generate_class($$) {
     my ($type, $isa) = @_;
+    my @isa;  # List::MoreUtils::uniq would go down a treat here, but let's try to keep this light on dependencies
+    my %seen; # we can't just map into a hash as we need to preserve the order
+
+    for my $superclass (@$isa) {
+        next if ($seen{$superclass});
+        push @isa, $superclass;
+        $seen{$superclass} = 1;
+    }
 
     # As an optimization, simply return the class if there's only one.
     # This speeds up method lookup as the method can (often) be found directly in the stash
     # rather than in the ISA hierarchy with its attendant AUTOLOAD-related overhead
-
-    if (@$isa == 1) {
-        my $class = $isa->[0];
+    if (@isa == 1) {
+        my $class = $isa[0];
         _universalize($class); # nop if it's already been universalized
         return $class;
     }
 
     # don't sort() the @isa as this incorrectly gives "use autobox SCALAR => [ qw(Foo Bar) ]" the same
     # synthetic class as "use autobox SCALAR => [ qw(Bar Foo) ]" - see isa.t
-    my $key = Storable::freeze([ $type, @$isa ]);
+    my $key = Storable::freeze([ $type, @isa ]);
 
     return $CLASS_CACHE->{$key} ||= do {
         my $class = sprintf('autobox::shim::<%d>', ++$SEQ);
         my $synthetic_class_isa = _get_isa($class); # i.e. autovivify
 
-        @$synthetic_class_isa = @$isa;
+        @$synthetic_class_isa = @isa;
         _universalize($class);
         $class;
     };
@@ -68,13 +75,15 @@ sub _universalize ($) {
 sub _annotate($) {
     my $hash = { %{ shift() } };
 
+    delete $hash->{_children}; # hide the housekeeping key
+
     # reverse() turns a hash that maps a type/isa signature to a class name into a hash that maps
     # a class name into a boolean
     my %synthetic = reverse(%$CLASS_CACHE);
 
     for my $key (keys %$hash) {
         my $value = $hash->{$key};
-        $hash->{$key} = join(', ', _get_isa($value)) if ($synthetic{$value});
+        $hash->{$key} = $synthetic{$value} ? [ _get_isa($value) ] : [ $value ];
     }
 
     return $hash;
@@ -82,7 +91,6 @@ sub _annotate($) {
 
 # default method called when the DEBUG option is supplied with a true value
 # prints the assigned bindings for the current scope
-
 sub _debug ($) {
     my $bindings = shift;
     require Data::Dumper;
@@ -119,23 +127,55 @@ sub _register ($) {
     $BINDINGS_CACHE->{$bindings} = $bindings; # keep the $bindings hash alive
 }
 
+# autobox.xs looks up the most specific type for each invocant e.g. INTEGER, FLOAT, STRING
+# these then need to be assigned bindings *if any of their supertypes have bindings*
+# this ensures that calling a method on e.g. a FLOAT delegates to NUMBER (if defined) then SCALAR
+# (if defined). This sub adds those implicit bindings
+
+sub _inherit($$@) {
+    my ($hash, $type, $supertype) = @_;
+
+    if (exists $hash->{$supertype}) { # don't autovivify
+        my $superclass = $hash->{$supertype};
+
+        if (exists $hash->{$type}) { # don't autovivify
+            $hash->{$type} = [ $hash->{$type} ] unless (_isa($hash->{$type}, 'ARRAY'));
+        } else {
+            $hash->{$type} = []; # autogenerate
+            # log the autogenerated subtypes so that we can remove them from the bindings 
+            # when the supertype is deleted
+            $hash->{_children}->{$supertype} ||= [];
+            push @{$hash->{_children}->{$supertype}}, $type;
+        }
+
+        push @{$hash->{$type}}, (_isa($superclass, 'ARRAY') ? @$superclass : $superclass);
+    }
+}
+
 ############################################# PUBLIC (Methods) ###############################################
 
 # allow subclasses to provide new defaults and/or allow a wider or narrower range of types to be autoboxed.
 #
 # the full list of supported types is:
 #
-#     REF SCALAR LVALUE ARRAY HASH CODE GLOB FORMAT IO UNKNOWN
+#     ARRAY BIND CODE FLOAT FORMAT GLOB HASH INTEGER IO LVALUE REF STRING UNDEF UNKNOWN VSTRING
+#
+# virtual types exist to assign classes to their subtypes and don't exist as types in their
+# own right
 
 sub defaults {
     my $class = shift;
     return {
-        SCALAR  => 'SCALAR',
         ARRAY   => 'ARRAY',
-        HASH    => 'HASH',
         CODE    => 'CODE',
-        UNDEF   => undef,
-        DEFAULT => undef
+        DEFAULT => undef,    # virtual
+        FLOAT   => undef,
+        HASH    => 'HASH',
+        INTEGER => undef,
+        NUMBER  => undef,    # virtual
+        SCALAR  => 'SCALAR', # virtual
+        STRING  => undef,
+        UNDEF   => undef
     };
 }
 
@@ -168,6 +208,7 @@ sub import {
     if (defined $default) {
         for my $key (keys %$defaults) {
             next if (exists $args{$key});    # don't supply a default if the binding is explicitly defined
+            next unless ($defaults->{$key}); # don't supply a default unless defaults() marks this type as defaultable
             next if ($key eq 'UNDEF');       # UNDEF must be autoboxed explicitly - DEFAULT doesn't include it
             next if ($key eq 'DEFAULT');     # already merged into $default above
 
@@ -196,9 +237,9 @@ sub import {
 
         # we can't use UNIVERSAL::isa to test if $value is an array ref;
         # if the value is 'ARRAY', and that package exists, then UNIVERSAL::isa('ARRAY', 'ARRAY') is true!
-
         $value = [ $value ] unless (_isa($value, 'ARRAY'));
 
+        # perform namespace expansion
         for my $class (@$value) {
             # squashed bug: if $class is a default namespace that was passed in an array ref,
             # then that array ref may have been assigned (above) as the default for multiple types
@@ -212,11 +253,43 @@ sub import {
             # see tests 27 and 28 in merge.t
             my $expanded = ($class =~ /::$/) ? "$class$type" : $class; # handle namespace expansion
 
-            push (@isa, $expanded) unless (grep { $_ eq $expanded } @isa); # no dups
+            push (@isa, $expanded); # dups are removed in _generate_class
         }
 
-        # associate the synthetic class with the specified type
-        $bindings->{$type} = _generate_class($type, \@isa);
+        $bindings->{$type} = [ @isa ]; # assign the namespace-expanded @isa to our $bindings hash
+    }
+
+    # thread together the class hierarchy for SCALAR subtypes; superclasses are only
+    # appended to the subtype's isa if bindings for the superclass are defined
+    # this must be done *after* namespace expansion and *before* _generate_class removes dups
+    # and collapses the @isa array ref into a class name
+    _inherit($bindings, 'NUMBER', 'SCALAR'); # the order is important
+    _inherit($bindings, 'STRING', 'SCALAR');
+    _inherit($bindings, 'INTEGER', 'NUMBER');
+    _inherit($bindings, 'FLOAT', 'NUMBER');
+
+    # finally: replace each array ref of superclasses with the name of the generated class.
+    # if there's only one superclass in the type's @ISA (e.g. SCALAR => 'MyScalar') then
+    # that class is used; otherwise a shim class whose @ISA contains the two or more classes
+    # is created 
+    for my $type (keys %$bindings) {
+        next if ($type eq '_children');
+
+        my $value = $bindings->{$type};
+
+        # the value may be a string inherited from the previous/outer scope
+        # there's no harm in resubmitting it to _generate_class: it
+        # will simply return the class name unchanged (from its cache)
+        
+        $value = [ $value ] unless (_isa($value, 'ARRAY'));
+
+        # delete empty arrays; empty arrays e.g. use autobox SCALAR => []
+        if (@$value == 0) {
+            delete $bindings->{$type};
+        } else {
+            # associate the synthetic/single class with the specified type
+            $bindings->{$type} = _generate_class($type, $value); 
+        }
     }
 
     # This turns on autoboxing i.e. the method call checker sets a flag on the method call op
@@ -236,7 +309,7 @@ sub import {
     # this is %^H as an integer - it changes as scopes are entered/exited
     # we don't need to stack/unstack it in %^H as %^H itself takes care of that
     # note: we need to call this *after* %^H is referenced, and possibly created, above
-    my $scope = Autobox::scope();
+    my $scope = autobox::scope();
     my $old_scope = exists($^H{autobox_scope})? $^H{autobox_scope} : 0;
     my $new_scope; # is this a new (top-level or nested) scope?
 
@@ -254,20 +327,20 @@ sub import {
     # This sub is called when this scope's $^H{autobox_leavescope} is deleted, usually when
     # %^H is destroyed at the end of the scope, but possibly directly in unimport()
     #
-    # Autobox::enterscope splices in the autobox method call checker and method call op if they're not already
+    # autobox::enterscope splices in the autobox method call checker and method call op if they're not already
     # active
     #
-    # Autobox::leavescope performs the necessary housekeeping to ensure that the default checker and op are restored
+    # autobox::leavescope performs the necessary housekeeping to ensure that the default checker and op are restored
     # when autobox is no longer in scope
 
     my $leave_scope = sub {
-        Autobox::leavescope();
+        autobox::leavescope();
     };
 
     my $guard = Scope::Guard->new($leave_scope);
     $^H{autobox_leavescope} = $guard;
 
-    Autobox::enterscope();
+    autobox::enterscope();
 }
 
 # delete one or more bindings; if none remain, disable autobox in the current scope
@@ -294,13 +367,30 @@ sub unimport {
     return unless ($^H{autobox});
 
     my $bindings = { %{$^H{autobox}} }; # clone the current bindings hash
+    my $children = exists $bindings->{_children} ? $bindings->{_children} : {};
 
     @args = keys(%$defaults) unless (@args);
 
-    for my $arg (@args) {
-        Carp::confess("unrecognized option: '", (defined $arg ? $arg : '<undef>'), "'")
-            unless (exists $defaults->{$arg});
-        delete $bindings->{$arg};
+    # cheap recursion; the real thing doesn't work because each frame creates its own copy of
+    # $bindings
+    while (@args) {
+        my $type = shift(@args);
+
+        Carp::confess("unrecognized option: '", (defined $type ? $type : '<undef>'), "'")
+            unless (exists $defaults->{$type});
+
+        # delete the children that were created automatically (i.e. not bound explicitly)
+        # for this type 
+        if (exists $children->{$type}) { # don't autovivify
+            push @args, @{$children->{$type}}; # add the children to the deletions todo list
+            delete $children->{$type};
+        }
+
+        delete $bindings->{$type}; # delete the specified type
+    }
+
+    unless (%$children) {
+        delete $bindings->{_children}; # harmless if it doesn't exist
     }
 
     if (%$bindings) {
